# vmaudit
# Author: Arcitec
# Project Site: https://github.com/Arcitec/vmaudit
# SPDX-License-Identifier: GPL-2.0-only


import sys
from enum import Enum
from typing import Union

# Global flag to control coloring.
ENABLE_COLOR = sys.stdout.isatty()


class Align(Enum):
    LEFT = "<"
    RIGHT = ">"
    CENTER = "^"


class Color(Enum):
    RESET = "\033[0m"
    BOLD = "\033[1m"

    RED = "\033[31m"
    GREEN = "\033[32m"
    YELLOW = "\033[33m"
    BLUE = "\033[34m"
    LIGHT_BLUE = "\033[94m"
    MAGENTA = "\033[35m"
    CYAN = "\033[36m"

    BOLD_YELLOW = "\033[1;33m"

    @classmethod
    def enable_color(cls, enabled: bool) -> None:
        """Globally enable or disable the ANSI colors."""
        global ENABLE_COLOR
        ENABLE_COLOR = enabled

    def __call__(self, text: Union[str, int]) -> str:
        """Return text wrapped in this ANSI terminal color (if TTY output)."""
        return f"{self.value}{text}{Color.RESET.value}" if ENABLE_COLOR else str(text)

    def pad(self, text: Union[str, int], padding: int = 0, align: Align = Align.RIGHT) -> str:
        """Return text wrapped in this ANSI terminal color (if TTY output), and padded to the chosen width."""
        if padding > 0:
            text = f"{text:{align.value}{padding}}"
        return self(text) if ENABLE_COLOR else str(text)



#!/usr/bin/env python3

# vmaudit
# Author: Arcitec
# Project Site: https://github.com/Arcitec/vmaudit
# SPDX-License-Identifier: GPL-2.0-only

# Entry point `__main__.py` is used when running via module name or directory:
#
# - `python -m vmaudit`
# - `python src/vmaudit`
#
# The installed `vmaudit` and `pipx run .` use `app.main` directly instead.
# NOTE: The latter methods verify `pyproject.toml` Python version requirements
# and dependencies during the installation, so we don't need to worry about that.


import sys

if sys.version_info < (3, 9):
    # Don't even attempt to load the program if they're running directly from
    # source code via an unsupported version of Python.
    print("ERROR: This program requires Python 3.9 or higher.")
    sys.exit(1)

if not __package__:
    # Inject packages with highest priority, to make source tree runnable via:
    # - `python src/vmaudit`
    # NOTE: This is not executed when running via `python -m vmaudit`.
    from pathlib import Path

    PACKAGES_DIR = Path(__file__).resolve().parent.parent
    sys.path.insert(0, str(PACKAGES_DIR))

from vmaudit.app import main

if __name__ == "__main__":
    sys.exit(main())



# vmaudit
# Author: Arcitec
# Project Site: https://github.com/Arcitec/vmaudit
# SPDX-License-Identifier: GPL-2.0-only


import math
from collections import defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Optional

from vmaudit.lib.cmdline import CmdLine, detect_app

# The kernel's default VMA limit is 65530, which is the max value that fits
# in an unsigned 16-bit integer (65535), minus 5 for some headroom. They
# chose that number because coredumps are written in ELF format, where the
# ELF "section" counter uses an unsigned 16-bit integer. So it's the highest
# amount of VMAs that can fit in a backwards-compatible ELF coredump.
# SEE: https://github.com/torvalds/linux/blob/v5.18/include/linux/mm.h#L178-L195
# NOTE: Higher values CAN be used, but will create a newer, "extended ELF" file.
KERNEL_DEFAULT_VMA_LIMIT = 65530


def get_vm_map_limit() -> int:
    """Return the maximum per-process VMA map limit used by the system."""
    try:
        limit_path = Path("/proc/sys/vm/max_map_count")
        limit = limit_path.read_text(encoding="utf-8").strip()
        if limit and limit.isdigit():
            return int(limit, 10)
    except (FileNotFoundError, PermissionError, ProcessLookupError):
        pass
    return 0


def get_vma_count(pid: int) -> int:
    """Return the number of VMAs (lines in `/proc/<pid>/maps`)."""
    # NOTE: The `maps` lines contain filenames and metadata, making it possible
    # to analyze exactly which shared `.so` libs and files are loaded.
    maps_path = Path(f"/proc/{pid}/maps")
    try:
        # Count newlines. Behavior matches "wc -l". Operates on 16x 4K blocks.
        with maps_path.open("rb") as f:  # Binary mode for speed.
            return sum(chunk.count(b"\n") for chunk in iter(lambda: f.read(65536), b""))
    except (FileNotFoundError, PermissionError, ProcessLookupError):
        return 0


@dataclass
class VMAreaStruct_Stats:
    active: int
    total: int


def read_slabinfo() -> Optional[VMAreaStruct_Stats]:
    """Try to read the Kernel's `/proc/slabinfo` and extract `vm_area_struct` stats."""
    # NOTE: Every VMA is allocated in the Kernel's slab cache, and is counted
    # in the slab cache's info, accessible via the following file.
    slabinfo_path = Path("/proc/slabinfo")
    if not slabinfo_path.is_file():
        return None

    try:
        with slabinfo_path.open("rt", encoding="utf-8") as f:
            for line in f:
                if line.startswith("vm_area_struct"):
                    parts = line.split()  # Split on consecutive whitespace runs.
                    if len(parts) >= 3:
                        # slabinfo format: name active_objs num_objs ...
                        return VMAreaStruct_Stats(active=int(parts[1], 10), total=int(parts[2], 10))
    except (PermissionError, FileNotFoundError):
        return None

    return None


@dataclass
class ProcessInfo:
    pid: int
    cmdline: CmdLine
    app_path: str
    vma_count: int


@dataclass
class VMA_ScanResults:
    processes: List[ProcessInfo]
    app_stats: Dict[str, List[int]]


def scan_vma_usage() -> VMA_ScanResults:
    """Return VMA statistics for all running processes."""
    processes: List[ProcessInfo] = []
    app_stats: Dict[str, List[int]] = defaultdict(list)

    for entry in Path("/proc").iterdir():
        if entry.name.isdigit():
            pid = int(entry.name, 10)
            vma_count = get_vma_count(pid)
            if vma_count > 0:
                cmdline = CmdLine.from_pid(pid)
                app_path = detect_app(cmdline)
                processes.append(ProcessInfo(pid, cmdline, app_path, vma_count))
                app_stats[app_path].append(vma_count)

    # Sort processes by highest VMA count (descending), then by name (ascending).
    # NOTE: The caller can easily get the "top process" since it'll be the 1st entry.
    processes.sort(key=lambda x: (-x.vma_count, x.app_path))

    return VMA_ScanResults(processes, app_stats)


@dataclass
class VMAOptimizer:
    """Performs analysis and optimization for the Linux kernels's VMA limit."""

    # Maximum number of VMAs observed or expected for a process.
    vma_peak: int

    def __post_init__(self) -> None:
        if self.vma_peak < 1:
            raise ValueError("VMA Peak must be a positive number.")

    def calculate_max_map_count(
        self, headroom_factor: float = 1.25, rounding_multiple: int = 10000
    ) -> int:
        """
        Calculates a recommended `vm.max_map_count` for the Linux kernel.

        That setting controls the maximum number of virtual memory areas (VMAs)
        a process can have. Applications that use many memory-mapped files or shared
        libraries, such as databases, search engines, or certain heavy Windows games,
        might need that limit increased.

        Note that the `vm.max_map_count` limit is just an integer, and isn't
        restricted to "powers of 2" or any other special rules.

        Args:
            headroom_factor: Multiplier for the safety margin. Defaults to 1.25 (25%).
            rounding_multiple: Produces nicely rounded recommendations. Set to 1 to
                disable. Defaults to 10000.

        Returns:
            The recommended integer value for the `vm.max_map_count` sysctl setting.
        """

        if headroom_factor < 1.2:
            raise ValueError(
                "Headroom factor should never be lower than 1.2 (20%), to ensure that processes have enough headroom for fluctuations."
            )

        if rounding_multiple < 1:
            raise ValueError("Rounding multiple must be 1 or higher.")

        # Determine the raw recommendation via the headroom factor.
        recommended = self.vma_peak * headroom_factor

        # Suggest the kernel's own default if the recommended number fits.
        if recommended <= KERNEL_DEFAULT_VMA_LIMIT:
            return KERNEL_DEFAULT_VMA_LIMIT
        else:
            # Round upwards to produce a nicer number.
            if rounding_multiple > 1:
                recommended = math.ceil(recommended / rounding_multiple) * rounding_multiple

            # Ensure it's never in the kernel's reserved 16-bit ELF header range.
            # NOTE: Replicates the kernel's "65535 - 5 reserved = 65530" behavior,
            # since other ELF data requires a few free slots in the 16-bit ELF range.
            if KERNEL_DEFAULT_VMA_LIMIT < recommended <= 65535:
                # Since the recommendation was in the reserved 16-bit ELF range, they
                # aren't using any rounding. So just push the recommendation above
                # a 16-bit number to force a kernel switch to "extended ELF" format.
                recommended = 65536

            return int(recommended)

    def calculate_vma_limit_headroom(self, vma_limit: int) -> float:
        """Return the remaining VMA limit headroom as a percentage."""
        return ((vma_limit - self.vma_peak) / vma_limit) * 100



# vmaudit
# Author: Arcitec
# Project Site: https://github.com/Arcitec/vmaudit
# SPDX-License-Identifier: GPL-2.0-only

import argparse
import getpass
import os
import re
import statistics
import sys
from typing import Optional, Union

from vmaudit.lib.cmdline import CmdLine
from vmaudit.lib.color import Color
from vmaudit.lib.vma import (
    KERNEL_DEFAULT_VMA_LIMIT,
    VMAOptimizer,
    get_vm_map_limit,
    read_slabinfo,
    scan_vma_usage,
)

ANONYMIZE = True

REAL_USERNAME = os.environ.get("SUDO_USER", getpass.getuser())

RGX_USERNAME = re.compile(re.escape(REAL_USERNAME), re.IGNORECASE)


def format_cmdline(cmdline: Union[CmdLine, str]) -> str:
    """Return a formatted and optionally anonymized command-line."""
    if isinstance(cmdline, CmdLine):
        cmdline = cmdline.to_string()

    if ANONYMIZE:
        # Very basic, but false positives don't matter here.
        # NOTE: Doesn't anonymize other names if running as root on multi-user
        # systems, but almost nobody uses personal computers like that.
        cmdline = RGX_USERNAME.sub("<user>", cmdline)

    return cmdline


def analyze_vma_usage(
    top_limit: int = 10,
    headroom_factor: float = 1.5,
    rounding_multiple: int = 10000,
    custom_vma_count: Optional[int] = None,
) -> None:
    """Analyze the system's current VMA usage and provide recommendations."""

    # Always show at least one detailed process (the heaviest process).
    if top_limit <= 0:
        top_limit = 1

    # Analyze all running processes.
    if os.geteuid() != 0:
        print(
            Color.RED(
                "** WARNING: This script is not running as root. "
                "You must run it as root if you want to scan all processes on the system. **"
            )
        )

    vma_stats = scan_vma_usage()

    if not vma_stats.processes:
        raise RuntimeError("No processes detected.")

    # The "VMA"-heaviest processes on the system.
    print(Color.LIGHT_BLUE(f"\n\n=== Top {top_limit} processes by VMA usage: ===\n"))
    print(
        Color.RED(
            "(WARNING: Do not share the detailed process top-list below. Their process arguments may contain SECRETS!)\n",
        )
    )
    print(
        f"{Color.BOLD_YELLOW.pad('VMA Count', 10)} "
        f"{Color.CYAN.pad('PID', 7)} "
        f"{Color.MAGENTA('Command')}"
    )
    print(Color.LIGHT_BLUE("-" * 80))

    for proc in vma_stats.processes[:top_limit]:
        print(
            f"{Color.BOLD_YELLOW.pad(proc.vma_count, 10)} "
            f"{Color.CYAN.pad(proc.pid, 7)} "
            f"{Color.MAGENTA(format_cmdline(proc.cmdline))}"
        )

    print(
        Color.RED(
            "\n(WARNING: Do not share the detailed process top-list above. Their process arguments may contain SECRETS!)",
        )
    )

    # Analyze per-application statistics.
    # NOTE: The kernel's VMA usage limit is per-process, not system-wide. Therefore,
    # the only thing that matters is the maximum VMA usage within a SINGLE process.
    # NOTE: When `fork()`-ing, child processes copy the parent's VMAs, but they
    # are independent COPIES and NOT counted as part of the parent's VMAs.
    # NOTE: Subprocess VMA usage does not count towards the parent's limit.
    print(Color.LIGHT_BLUE("\n\n=== Per-application VMA statistics: ===\n"))
    print(
        f"{Color.CYAN.pad('Instances', 10)} "
        f"{Color.BOLD_YELLOW.pad('Max VMAs', 10)} "
        f"{Color.GREEN.pad('Avg VMAs', 10)} "
        f"{Color.GREEN.pad('Min VMAs', 10)} "
        f"{Color.MAGENTA('Application')}"
    )
    print(Color.LIGHT_BLUE("-" * 80))

    for app_name, instance_vmas in sorted(
        vma_stats.app_stats.items(),
        # Sort apps by highest VMA count (descending), then by name (ascending).
        # NOTE: There's always at least one app-instance, so the list is never empty.
        key=lambda kv: (-max(kv[1]), kv[0]),
    ):
        instances = len(instance_vmas)
        app_vma_max = max(instance_vmas)
        app_vma_avg = round(statistics.mean(instance_vmas))
        app_vma_min = min(instance_vmas)

        print(
            f"{Color.CYAN.pad(instances, 10)} "
            f"{Color.BOLD_YELLOW.pad(app_vma_max, 10)} "
            f"{Color.GREEN.pad(app_vma_avg, 10)} "
            f"{Color.GREEN.pad(app_vma_min, 10)} "
            f"{Color.MAGENTA(format_cmdline(app_name))}"
        )

    total_vmas = sum(proc.vma_count for proc in vma_stats.processes)

    print(Color.LIGHT_BLUE("-" * 80))
    print(f"Total VMA count across all detected processes: {Color.BOLD_YELLOW(total_vmas)}")

    # Analyze the kernel's true VMA metadata count and limit.
    print(Color.LIGHT_BLUE("\n\n=== Kernel VMA metadata statistics: ===\n"))
    slabinfo = read_slabinfo()
    slabinfo_msg = (
        f"active={Color.BOLD_YELLOW(slabinfo.active)}, total={Color.BOLD_YELLOW(slabinfo.total)}"
        if slabinfo
        else "slabinfo not available (need to run as root to see kernel statistics)"
    )
    print(f"{Color.CYAN.pad('vm_area_struct:', 18)}  {slabinfo_msg}")

    vm_map_limit = get_vm_map_limit()
    print(f"{Color.CYAN.pad('vm.max_map_count:', 18)}  limit={Color.BOLD_YELLOW(vm_map_limit)}\n")

    print(
        f"(INFO: The kernel VMA max limit is {Color.CYAN('per-process')}. "
        f"The system's combined, {Color.CYAN('total')} VMA count {Color.CYAN('does not matter')}.)"
    )

    # Provide recommendations for tuning the VMA limit.
    print(Color.LIGHT_BLUE("\n\n=== Kernel VMA tuning recommendations: ===\n"))
    top_process = vma_stats.processes[0]

    add_divider = False
    for process_title, vma_count in [
        (top_process.app_path, top_process.vma_count),
        ("custom --vma value", custom_vma_count),
    ]:
        if vma_count is None:
            continue

        if add_divider:
            print(Color.LIGHT_BLUE(f"{'-' * 80}\n"))
        add_divider = True

        print(f"{Color.CYAN.pad('top process:', 18)}  {Color.MAGENTA(process_title)}")
        print(f"{Color.CYAN.pad('VMA usage:', 18)}  {Color.BOLD_YELLOW(vma_count)}\n")

        vma_optimizer = VMAOptimizer(vma_count)

        recommended_limit = vma_optimizer.calculate_max_map_count(
            headroom_factor, rounding_multiple
        )

        for title, limit in [
            ("current limit", vm_map_limit),
            ("recommended", recommended_limit),
        ]:
            headroom_percent = vma_optimizer.calculate_vma_limit_headroom(limit)
            if headroom_percent >= 20.0:
                headroom_color = Color.GREEN
            elif headroom_percent >= 5.0:
                headroom_color = Color.YELLOW
            else:
                headroom_color = Color.RED

            title_color = Color.GREEN if title == "recommended" else Color.CYAN

            print(
                f"{title_color.pad(f'{title}:', 18)}  vm.max_map_count={Color.BOLD_YELLOW(limit)} "
                f"({'kernel default' if limit == KERNEL_DEFAULT_VMA_LIMIT else 'custom value'})"
            )
            print(
                f"{Color.CYAN.pad('headroom:', 18)}  free VMAs={Color.BOLD_YELLOW(limit - vma_count)}, "
                f"free percent={headroom_color(f'{headroom_percent:.1f}%')}\n"
            )


def main() -> int:
    parser = argparse.ArgumentParser(
        prog="vmaudit",
        description="Analyzes Virtual Memory Area (VMA) utilization to provide data-driven recommendations for system administrators and developers.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    parser.add_argument("--top", type=int, default=20, help="how many processes to display")

    # Use a very generous, overkill "50% additional headroom" recommendation
    # to allow huge, unpredictable fluctuations. If a workload is predictable,
    # a headroom of "25%" is more reasonable.
    parser.add_argument("--headroom", type=float, default=1.5, help="headroom multiplier")

    parser.add_argument("--round", type=int, default=10000, help="headroom rounding multiple")

    # Allow the user to specify a custom VMA value to also calculate.
    parser.add_argument(
        "--vma", type=int, help="optional: largest expected VMA usage for a process"
    )

    args = parser.parse_args()

    try:
        analyze_vma_usage(args.top, args.headroom, args.round, args.vma)
    except Exception as e:
        # Detect the file and line number that threw the exception.
        exc_type, exc_value, exc_traceback = sys.exc_info()
        last_frame = exc_traceback
        while last_frame is not None and last_frame.tb_next:
            last_frame = last_frame.tb_next
        line_number = last_frame.tb_lineno if last_frame else -1
        file_name = last_frame.tb_frame.f_code.co_filename if last_frame else "?"

        # Show a simplified error message with enough information.
        print(Color.RED(f'ERROR: ["{os.path.basename(file_name)}:{line_number}"] {e}'))

        return 1

    return 0


if __name__ == "__main__":
    sys.exit(main())



# vmaudit
# Author: Arcitec
# Project Site: https://github.com/Arcitec/vmaudit
# SPDX-License-Identifier: GPL-2.0-only


import re
import shlex
from dataclasses import dataclass
from pathlib import Path
from typing import List


@dataclass
class CmdLine:
    args: List[str]

    def to_string(self) -> str:
        """Return the command-line as a shell-escaped string."""
        return shlex.join(self.args)

    @classmethod
    def from_pid(cls, pid: int) -> "CmdLine":
        """Get the command-line for the given process."""
        proc_dir = Path(f"/proc/{pid}")

        try:
            cmdline_path = proc_dir / "cmdline"
            raw = cmdline_path.read_bytes()  # Bytes, usually NUL-separated args.
            # Split into bytes-chunks on NUL, decode and strip each part.
            parts: List[str] = [
                part.decode("utf-8", errors="replace").strip()
                for part in raw.split(b"\x00")
                if part  # Skip empty bytes chunks quickly.
            ]
            if parts:
                # Handle special case: Uses spaces instead of NUL-separated args.
                # NOTE: Can happen with some containerized processes.
                # NOTE: The processing is nasty since there's *nothing* that tells
                # us where the command ends and the arguments begin.
                if len(parts) == 1:
                    # First attempt to split on Windows ".exe", so that we never
                    # end up misdetecting Windows executable paths.
                    # NOTE: If there's nothing after the divider, we get an empty
                    # second element. But at least we've detected that it's an exe.
                    arg_split: List[str]
                    is_windows = re.search(r"^[a-zA-Z]:[\/\\]", parts[0]) is not None
                    if is_windows:
                        # Only attempt ".exe" split if path began with drive letter.
                        arg_split = re.split(r"(?i)(\.exe)\b\s*", parts[0], maxsplit=1)
                        if len(arg_split) == 3:
                            # When splitting with a capture, it captures the divider
                            # as a separate entry, so we need to join them again.
                            arg_split = [f"{arg_split[0]}{arg_split[1]}", arg_split[2]]
                    if not is_windows or len(arg_split) != 2:
                        # If we didn't find a ".exe", attempt to split on earliest
                        # "[whitespace][dash][not whitespace]" to find arguments.
                        # NOTE: There's no better way to do this. The command
                        # is stored without any quoted paths, so we can't tell the
                        # difference between a flag or a directory named "/a - b/"
                        # in the path, so the best we can do is restrict it to
                        # flags that are followed by a non-whitespace character.
                        # NOTE: This will have false negatives whenever the argument
                        # is not preceded by a "-", but there's no way solve that.
                        # NOTE: We can't split via shlex directly, since it would
                        # instantly split on the first whitespace, which is wrong.
                        arg_split = re.split(r"\s+(-\S)", parts[0], maxsplit=1)
                        if len(arg_split) == 3:
                            # Join the split-divider with the arguments again.
                            arg_split = [arg_split[0], f"{arg_split[1]}{arg_split[2]}"]

                    # If we've split on ".exe" or arguments, post-process the args.
                    if len(arg_split) == 2:
                        if len(arg_split[1]) == 0:
                            # Remove empty trailing split element.
                            arg_split.pop()
                        else:
                            # Post-process the argument-string via shlex to split
                            # into individual arguments.
                            # NOTE: Shlex really hates Windows backslashes in paths
                            # and erases them, which happens whenever Wine apps
                            # use Windows paths in any of their arguments. :/
                            try:
                                extra_args = shlex.split(arg_split[1])
                                # Replace the extra arguments with the shlex split.
                                arg_split.pop()
                                arg_split.extend(extra_args)
                            except ValueError:
                                # Fallback: Keep as single argument if shlex fails.
                                pass

                        # Place the final command and args in the list.
                        parts = arg_split

                return cls(parts)
        except (FileNotFoundError, PermissionError, ProcessLookupError):
            pass

        # Fallback to "comm" parsing (single command, no arguments).
        try:
            comm_path = proc_dir / "comm"
            comm = comm_path.read_text(encoding="utf-8").strip()
            if comm:
                return cls([comm])
        except (FileNotFoundError, PermissionError, ProcessLookupError):
            pass

        return cls([f"[pid {pid}]"])


def detect_app(cmdline: CmdLine) -> str:
    """
    Detect the "real" application from complex command-line arguments.

    The parsing is pretty basic, but good enough for the purpose.
    """

    args = cmdline.args

    # Ignore the bwrap Flatpak sandboxing wrapper.
    if args and args[0] == "/usr/bin/bwrap":
        try:
            bwrap_cmdsep = args.index("--")  # Separated by double dash.
            args = args[bwrap_cmdsep + 1 :]
        except ValueError:
            # No "--" found; keep as-is.
            pass

    # Ignore the zypak Chromium/Electron Flatpak integration helper.
    if args and args[0] == "/app/bin/zypak-helper":
        try:
            zypak_cmdsep = args.index("-")  # Separated by single dash.
            args = args[zypak_cmdsep + 1 :]
        except ValueError:
            # No "-" found; keep as-is.
            pass

    # Ignore the Python2/3 interpreter and detect actual script name instead.
    if args and re.search(r"(?:^|\/)python[\d.]*$", args[0]):
        # Attempt to find the first non-flag Python argument.
        for i, arg in enumerate(args[1:], start=1):
            if not arg.startswith("-"):
                args = args[i:]
                if "\n" in args[0]:
                    # Python has been executed with "-c" code. Only keep line 1.
                    py_code = args[0].splitlines()
                    return f'[Python Code "{py_code[0].strip()}"...]'
                break

    if not args:
        return "[unknown]"

    # Detect a very common Electron subprocess wrapper.
    is_electron_subproc = args[0] == "/proc/self/exe"

    # Analyze the individual arguments/tokens to learn more about the process.
    for t in args:
        if ".asar" in t:
            # Handle "--flag=/path/to/app.asar".
            if t.startswith("--") and "=" in t:
                after_eq = t.split("=", 1)[1]  # Empty "" if nothing after "=".
                if ".asar" in after_eq:
                    t = after_eq
            # It's an Electron ".asar" app package.
            return f'[Electron App "{t}"]'
        elif is_electron_subproc and t.startswith("--user-data-dir="):
            # Anonymous Electron child process. Identify via its data dir.
            return f'[Electron Helper for "{t[16:]}"]'

    # Use the first argument (the Linux/Wine application path).
    app_name = args[0]
    if re.search(r"\.exe$", app_name, re.IGNORECASE):
        return f'[Wine App "{app_name}"]'
    return app_name
